import asyncio
import json
import tempfile

from linty_fresh.problem import Problem

from typing import Set


MAX_REVISIONS = 10
NOTES_REF = 'refs/notes/linty_fresh'
GIT_SUBPROCESS_KWARGS = {
    'stdout': asyncio.subprocess.PIPE,
    'stderr': asyncio.subprocess.DEVNULL,
    'env': {
        'GIT_NOTES_REF': NOTES_REF
    }
}


class GitNotesStorageEngine(object):
    def __init__(self, remote: str=None):
        self.remote = remote

    async def get_existing_problems(self) -> Set[Problem]:
        note_ref = ''
        result = set()
        if self.remote:
            fetch_notes = await asyncio.create_subprocess_exec(
                'git', 'fetch', self.remote, '{0}:{0}'.format(NOTES_REF),
                **GIT_SUBPROCESS_KWARGS)
            await fetch_notes.wait()

        last_n_revisions_proc = await asyncio.create_subprocess_exec(
            'git', 'log', '--skip=1', '-{}'.format(MAX_REVISIONS),
            '--pretty=%H',
            **GIT_SUBPROCESS_KWARGS)
        async for line in last_n_revisions_proc.stdout:
            notes_ref_proc = await asyncio.create_subprocess_exec(
                'git', 'notes', 'list', line.decode().strip(),
                **GIT_SUBPROCESS_KWARGS)
            await notes_ref_proc.wait()
            if notes_ref_proc.returncode == 0:
                note_ref = (
                    await notes_ref_proc.stdout.readline()).decode().strip()
                if note_ref:
                    break
        await last_n_revisions_proc.wait()
        if note_ref:
            notes_proc = await asyncio.create_subprocess_exec(
                'git', 'show', note_ref,
                **GIT_SUBPROCESS_KWARGS)
            async for line in notes_proc.stdout:
                try:
                    problems = json.loads(line.decode())
                    for problem in problems:
                        result.add(Problem.from_json(problem))
                except Exception:
                    pass
            await notes_proc.wait()
        return result

    async def store_problems(self, problems: Set[Problem]) -> None:
        with tempfile.NamedTemporaryFile() as problem_file:
            result = sorted([problem.to_json() for problem in problems],
                            key=lambda x: str(x))
            problem_file.write(json.dumps(result).encode())
            problem_file.flush()
            notes_proc = await asyncio.create_subprocess_exec(
                'git', 'notes', 'append', '-F', problem_file.name,
                **GIT_SUBPROCESS_KWARGS)
            await notes_proc.wait()
            if self.remote:
                push_proc = await asyncio.create_subprocess_exec(
                    'git', 'push', '-f', '-q', self.remote, NOTES_REF,
                    **GIT_SUBPROCESS_KWARGS)
                await push_proc.wait()



import re

from linty_fresh.problem import Problem

from typing import Set


MYPY_LINE_REGEX = re.compile(r'(?P<path>[^:]*):(?P<line>\d*):\s*'
                             r'(?P<code>\w*)\s*:\s*(?P<message>.*)')


def parse(contents: str) -> Set[Problem]:
    result = set()  # type: Set[Problem]
    for line in contents.splitlines():
        match = MYPY_LINE_REGEX.match(line)
        if match:
            groups = match.groupdict()
            if groups['code'] != 'note':
                result.add(Problem(groups['path'],
                                   groups['line'],
                                   '{}: {}'.format(groups['code'],
                                                   groups['message'])))
    return result



import os
import re

from linty_fresh.problem import Problem

from typing import Set

SWIFTLINT_LINE_REGEX = re.compile(r'(?P<path>[^:]*):(?P<line>\d*):'
                                  r'(?:(?P<column>\d*):)?\s*(?P<level>\w*):'
                                  r'\s*(?P<message>.*)')


def parse(contents: str) -> Set[Problem]:
    result = set()  # type: Set[Problem]
    for line in contents.splitlines():
        match = SWIFTLINT_LINE_REGEX.match(line)
        if match:
            groups = match.groupdict()
            result.add(Problem(os.path.relpath(groups['path']),
                               groups['line'],
                               groups['message']))
    return result



import argparse
import asyncio

from linty_fresh.linters import checkstyle, mypy, pylint, swiftlint
from linty_fresh.reporters import github_reporter

from typing import Any, Dict  # noqa

from linty_fresh.storage.git_storage_engine import GitNotesStorageEngine

REPORTERS = {
    'github': github_reporter,
}  # type: Dict[str, Any]

LINTERS = {
    'mypy': mypy,
    'pylint': pylint,
    'swiftlint': swiftlint,
    'checkstyle': checkstyle,
}  # type: Dict[str, Any]


def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument('--reporter', type=str, nargs='+', default=['github'],
                        help='The reporter to use when reporting errors.')
    parser.add_argument('--linter', type=str, required=True,
                        help='The type of lint file to parse.')
    parser.add_argument('--store_problems', default=False, action='store_true',
                        help='Whether or not to store lint errors as git '
                             'notes and filter out problems')
    parser.add_argument('files', type=str, nargs='+',
                        help='The lint file being parsed.')
    for name, reporter in REPORTERS.items():
        reporter.register_arguments(parser)
    return parser


async def run_loop(args):
    args = create_parser().parse_args()
    reporters = []
    for reporter in args.reporter:
        if reporter not in REPORTERS:
            raise Exception('Reporter {} is invalid, options are {}'.format(
                reporter, ','.join(list(REPORTERS.keys()))
            ))
        reporters.append(REPORTERS[reporter].create_reporter(args))

    problems = set()
    if args.linter not in LINTERS:
        raise Exception('Linter {} is invalid, options are {}'.format(
            args.linter, ','.join(list(LINTERS.keys()))
        ))
    linter = LINTERS[args.linter]
    for lint_file_path in args.files:
        with open(lint_file_path, 'r') as lint_file:
            problems.update(linter.parse(lint_file.read()))

    storage_engine = GitNotesStorageEngine('origin')

    awaitable_array = []

    if args.store_problems:
        awaitable_array.append(storage_engine.store_problems(problems))
        existing_problems = await storage_engine.get_existing_problems()
        problems = problems.difference(existing_problems)

    awaitable_array.extend([reporter.report(problems) for
                            reporter in
                            reporters])
    await asyncio.gather(*awaitable_array)


def main():
    args = create_parser().parse_args()
    loop = asyncio.get_event_loop()
    loop.run_until_complete(run_loop(args))
    loop.close()


if __name__ == '__main__':
    main()



import argparse
import asyncio
import json
import os
import re

from collections import defaultdict

import aiohttp

from linty_fresh.problem import Problem

from typing import Any, Dict, List, MutableMapping, NamedTuple, Set


PR_URL_REGEX = re.compile(r'https?://.*?github.com/'
                          r'(?:repos/)?'
                          r'(?P<organization>[^/]*)/'
                          r'(?P<repo>[^/]*)/pulls?/'
                          r'(?P<pr_number>\d*)')
HUNK_REGEX = re.compile(r'@@ \-\d+,\d+ \+(\d+),\d+ @@')
FILE_START_REGEX = re.compile(r'\+\+\+ b/(.*)')
LINK_REGEX = re.compile(r'<(?P<url>.+)>; rel="(?P<rel>\w+)"')
NEW_FILE_SECTION_START = 'diff --git a'
MAX_LINT_ERROR_REPORTS = 10

ExistingGithubMessage = NamedTuple('ExistingGithubMessage',
                                   [('path', str),
                                    ('position', int),
                                    ('body', str)])


class GithubReporter(object):
    def __init__(self,
                 auth_token: str,
                 organization: str,
                 repo: str,
                 pr_number: int,
                 commit: str) -> None:
        self.auth_token = auth_token
        self.organization = organization
        self.repo = repo
        self.pr = pr_number
        self.commit = commit

    async def report(self, problems: List[Problem]) -> None:
        grouped_problems = Problem.group_by_path_and_line(problems)

        headers = {
            'Authorization': 'token {}'.format(self.auth_token),
        }
        with aiohttp.ClientSession(headers=headers) as client_session:
            (line_map, existing_messages) = await asyncio.gather(
                self.create_line_to_position_map(client_session),
                self.get_existing_messages(client_session))
            lint_errors = 0
            review_comment_awaitable = []
            pr_url = self._get_pr_url()
            no_matching_line_number = []
            for location, problems_for_line in grouped_problems:
                message_for_line = [':sparkles:Linty Fresh Says:sparkles::',
                                    '']

                reported_problems_for_line = set()

                path = location[0]
                line_number = location[1]
                position = line_map.get(path, {}).get(line_number, None)
                if position is None and path in line_map:
                    file_map = line_map[path]
                    closest_line = min(file_map.keys(),
                                       key=lambda x: abs(x-line_number))
                    position = file_map[closest_line]
                    message_for_line.append('(From line {})'.format(
                        line_number))
                message_for_line.append('```')
                if position is not None:
                    for problem in problems_for_line:
                        if problem.message not in reported_problems_for_line:
                            message_for_line.append(problem.message)
                            reported_problems_for_line.add(problem.message)
                    message_for_line.append('```')
                    message = '\n'.join(message_for_line)
                    if (path, position, message) not in existing_messages:
                        lint_errors += 1
                        if lint_errors <= MAX_LINT_ERROR_REPORTS:
                            data = json.dumps({
                                'body': message,
                                'commit_id': self.commit,
                                'path': path,
                                'position': position,
                            }, sort_keys=True)
                            review_comment_awaitable.append(
                                client_session.post(pr_url, data=data))
                else:
                    no_matching_line_number.append((location,
                                                    problems_for_line))
            if lint_errors > MAX_LINT_ERROR_REPORTS:
                message = ''':sparkles:Linty Fresh Says:sparkles::

Too many lint errors to report inline!  {0} lines have a problem.
Only reporting the first {1}.'''.format(
                    lint_errors, MAX_LINT_ERROR_REPORTS)
                data = json.dumps({
                    'body': message
                })
                review_comment_awaitable.append(
                    asyncio.ensure_future(client_session.post(
                        self._get_issue_url(),
                        data=data)))
            if no_matching_line_number:
                no_matching_line_messages = []
                for location, problems_for_line in no_matching_line_number:
                    path = location[0]
                    line_number = location[1]
                    no_matching_line_messages.append(
                        '{0}:{1}:'.format(path, line_number))
                    for problem in problems_for_line:
                        no_matching_line_messages.append('\t{0}'.format(
                            problem.message))
                message = ('Linters found some problems with lines not '
                           'modified by this commit:\n```\n{0}\n```'.format(
                               '\n'.join(no_matching_line_messages)))
                data = json.dumps({
                    'body': message
                })
                review_comment_awaitable.append(
                    asyncio.ensure_future(client_session.post(
                        self._get_issue_url(), data=data)))

            responses = await asyncio.gather(
                *review_comment_awaitable
            )  # type: List[aiohttp.ClientResponse]
            for response in responses:
                response.close()

    async def create_line_to_position_map(
        self, client_session: aiohttp.ClientSession
    ) -> MutableMapping[str, Dict[int, int]]:
        result = defaultdict(dict)  # type: MutableMapping[str, Dict[int, int]]
        current_file = ''
        position = -1
        right_line_number = -1

        headers = {
            'Accept': 'application/vnd.github.diff',
        }
        url = ('https://api.github.com/repos/'
               '{organization}/{repo}/pulls/{pr}'.format(
                   organization=self.organization,
                   repo=self.repo,
                   pr=self.pr))
        async with client_session.get(url, headers=headers) as response:
            async for line in response.content:
                line = line.decode()
                file_match = FILE_START_REGEX.match(line)
                if file_match:
                    current_file = file_match.groups()[0].strip()
                    right_line_number = -1
                    position = -1
                    continue
                elif line.startswith(NEW_FILE_SECTION_START):
                    current_file = ''
                if not current_file:
                    continue

                position += 1
                hunk_match = HUNK_REGEX.match(line)
                if hunk_match:
                    right_line_number = int(hunk_match.groups()[0]) - 1
                elif not line.startswith('-'):
                    right_line_number += 1
                    result[current_file][right_line_number] = position

        return result

    def _get_pr_url(self) -> str:
        return ('https://api.github.com/repos/'
                '{organization}/{repo}/pulls/{pr}/comments'.format(
                    organization=self.organization,
                    repo=self.repo,
                    pr=self.pr))

    def _get_issue_url(self) -> str:
        return ('https://api.github.com/repos/'
                '{organization}/{repo}/issues/{pr}/comments'.format(
                    organization=self.organization,
                    repo=self.repo,
                    pr=self.pr))

    async def get_existing_messages(
        self, client_session: aiohttp.ClientSession
    ) -> Set[ExistingGithubMessage]:
        url = self._get_pr_url()
        return await self._fetch_messages_from_url(client_session, url)

    @staticmethod
    async def _fetch_messages_from_url(
            client_session, url
    ) -> Set[ExistingGithubMessage]:
        existing_messages = set()  # type: Set[ExistingGithubMessage]
        async with client_session.get(url) as response:
            response = response  # type: aiohttp.ClientResponse
            comments = json.loads(await response.text())
            for comment in comments:
                existing_messages.add(
                    ExistingGithubMessage(path=comment['path'],
                                          position=comment['position'],
                                          body=comment['body']))

            next_url = GithubReporter._find_next_url(response)

        if next_url:
            existing_messages = existing_messages.union(
                await GithubReporter._fetch_messages_from_url(
                    client_session, next_url))

        return existing_messages

    @staticmethod
    def _find_next_url(response: aiohttp.ClientResponse) -> str:
        if 'link' in response.headers:
            links = response.headers['link'].split(',')
            for link in links:
                match = LINK_REGEX.match(link)
                if match and match.group('rel') == 'next':
                    return match.group('url')


def register_arguments(parser: argparse.ArgumentParser) -> None:
    parser.add_argument('--pr_url',
                        type=str,
                        help='The URL for the Pull Request for this commit.')
    parser.add_argument('--commit',
                        type=str,
                        help='The commit being linted.')


def create_reporter(args: Any) -> GithubReporter:
    if not args.pr_url or not args.commit:
        raise Exception('Must specify both a pr_url and a commit to use the '
                        'github reporter.')
    auth_token = os.getenv('GITHUB_AUTH_TOKEN')
    if not auth_token:
        raise Exception('Environment Variable $GITHUB_AUTH_TOKEN must be set '
                        'to use the github reporter.')
    match = PR_URL_REGEX.match(args.pr_url)
    if match:
        groups = match.groupdict()
        return GithubReporter(auth_token,
                              groups['organization'],
                              groups['repo'],
                              int(groups['pr_number']),
                              args.commit)
    else:
        raise Exception("{} doesn't appear to be a valid github pr url".format(
            args.pr_url
        ))



from xml.etree import ElementTree

from linty_fresh.problem import Problem

from typing import Set


def parse(contents: str) -> Set[Problem]:
    result = set()  # type: Set[Problem]
    try:
        root = ElementTree.fromstring(contents)
    except ElementTree.ParseError:
        return result
    for file in root.findall('file'):
        file_name = file.get('name')
        for error in file.findall('error'):
            result.add(Problem(
                file_name,
                error.get('line'),
                '{}: {}'.format(
                    error.get('source'),
                    error.get('message'))))
    return result



from setuptools import find_packages, setup


def get_version(file_name='linty_fresh/__init__.py'):
    with open(file_name) as f:
        for line in f.read().splitlines():
            if line.startswith('__version__'):
                return line.split('=')[-1].strip(" '")

setup(
    name='linty-fresh',
    version=get_version(),
    license='apache',
    description='''
        This package reports style violations for a Github PR as comments.
    ''',
    author='Roy Williams',
    author_email='rwilliams@lyft.com',
    url='https://github.com/lyft/linty-fresh',
    packages=find_packages(exclude=['tests*']),
    include_package_data=True,
    setup_requires=[
        'flake8>=2.2.3',
        'flake8-blind-except>=0.1.0',
        'flake8-debugger>=1.4.0',
        'flake8-import-order>=0.6.1',
        'flake8-quotes>=0.1.1',
        'pep8==1.5.7',
        'pep8-naming>=0.3.3',
    ],
    install_requires=[
        'aiohttp>=0.19.0',
        'typing>=3.5.0.1',
    ],
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Quality Assurance',
    ],
    entry_points={
        'console_scripts': ['linty_fresh = linty_fresh.main:main'],
    },
    tests_require=[
        'nose>=1.3.3',
        'mock>=1.0.1',
    ],
    test_suite='nose.collector',
)



import re

from linty_fresh.problem import Problem

from typing import Set

PYLINT_LINE_REGEX = re.compile(r'(?P<path>[^:]*):(?P<line>\d*):\s*'
                               r'(?:(?P<column>\d*):)?\s*'
                               r'(?P<message>.*)')


def parse(contents: str) -> Set[Problem]:
    result = set()  # type: Set[Problem]
    for line in contents.splitlines():
        match = PYLINT_LINE_REGEX.match(line)
        if match:
            groups = match.groupdict()
            result.add(Problem(groups['path'],
                               groups['line'],
                               groups['message']))
    return result



import itertools

from collections import OrderedDict


class Problem(object):
    def __init__(self, file_path: str, line: str, message: str) -> None:
        self.path = file_path
        self.line = int(line)
        self.message = message

    def __hash__(self):
        return hash((self.path, self.line, self.message))

    def __eq__(self, other):
        return ((self.path, self.line, self.message) ==
                (other.path, other.line, other.message))

    def to_json(self):
        return OrderedDict({
            'path': self.path,
            'line': self.line,
            'message': self.message
        })

    @staticmethod
    def from_json(json_object) -> 'Problem':
        return Problem(json_object['path'],
                       json_object['line'],
                       json_object['message'])

    @staticmethod
    def group_by_path_and_line(problems):
        def key_func(problem):
            return problem.path, problem.line

        sorted_problems = sorted(problems, key=key_func)
        return [(location, list(problems)) for (location, problems) in
                itertools.groupby(sorted_problems, key_func)]
